@@ -33,7 +33,7 @@
 #define NUMERICOID 1700
 
 
-void pktbuf_free(PktBuf *buf)
+static void pktbuf_free_internal(PktBuf *buf)
 {
 	if (!buf || buf->fixed_buf)
 		return;
@@ -44,6 +44,20 @@ void pktbuf_free(PktBuf *buf)
 	free(buf);
 }
 
+static PktBuf *temp_pktbuf;
+
+/*
+ * free the given buffer, if it's a dynamic buffer and not the global temp
+ * buffer
+ */
+void pktbuf_free(PktBuf *buf)
+{
+	if (buf == temp_pktbuf)
+		return;
+
+	pktbuf_free_internal(buf);
+}
+
 PktBuf *pktbuf_dynamic(int start_len)
 {
 	PktBuf *buf = zmalloc(sizeof(PktBuf));
@@ -82,8 +96,6 @@ void pktbuf_static(PktBuf *buf, uint8_t *data, int len)
 	buf->fixed_buf = true;
 }
 
-static PktBuf *temp_pktbuf;
-
 struct PktBuf *pktbuf_temp(void)
 {
 	if (!temp_pktbuf)
@@ -96,7 +108,7 @@ struct PktBuf *pktbuf_temp(void)
 
 void pktbuf_cleanup(void)
 {
-	pktbuf_free(temp_pktbuf);
+	pktbuf_free_internal(temp_pktbuf);
 	temp_pktbuf = NULL;
 }
 
@@ -241,6 +253,9 @@ void pktbuf_put_uint64(PktBuf *buf, uint64_t val)
 
 void pktbuf_put_bytes(PktBuf *buf, const void *data, int len)
 {
+	if (len == 0)
+		return;
+
 	make_room(buf, len);
 	if (buf->failed)
 		return;
@@ -448,8 +463,7 @@ void pktbuf_write_DataRow(PktBuf *buf, const char *tupdesc, ...)
 				for (int j = 0; j < blen; j++)
 					sprintf(tmp + (2 + j * 2), "%02x", bval[j]);
 				val = tmp;
-			}
-			else {
+			} else {
 				(void) va_arg(ap, uint8_t *);
 				val = NULL;
 			}
