@@ -25,28 +25,42 @@ static void reset_stats(PgStats *stat)
 {
 	stat->server_bytes = 0;
 	stat->client_bytes = 0;
+	stat->server_assignment_count = 0;
 	stat->query_count = 0;
 	stat->query_time = 0;
 	stat->xact_count = 0;
 	stat->xact_time = 0;
 	stat->wait_time = 0;
+
+	stat->ps_client_parse_count = 0;
+	stat->ps_server_parse_count = 0;
+	stat->ps_bind_count = 0;
 }
 
 static void stat_add(PgStats *total, PgStats *stat)
 {
 	total->server_bytes += stat->server_bytes;
 	total->client_bytes += stat->client_bytes;
+	total->server_assignment_count += stat->server_assignment_count;
 	total->query_count += stat->query_count;
 	total->query_time += stat->query_time;
 	total->xact_count += stat->xact_count;
 	total->xact_time += stat->xact_time;
 	total->wait_time += stat->wait_time;
+
+	total->ps_client_parse_count += stat->ps_client_parse_count;
+	total->ps_server_parse_count += stat->ps_server_parse_count;
+	total->ps_bind_count += stat->ps_bind_count;
 }
 
 static void calc_average(PgStats *avg, PgStats *cur, PgStats *old)
 {
+	uint64_t server_assignment_count;
 	uint64_t query_count;
 	uint64_t xact_count;
+	uint64_t ps_client_parse_count;
+	uint64_t ps_server_parse_count;
+	uint64_t ps_bind_count;
 
 	usec_t dur = get_cached_time() - old_stamp;
 
@@ -57,9 +71,11 @@ static void calc_average(PgStats *avg, PgStats *cur, PgStats *old)
 
 	query_count = cur->query_count - old->query_count;
 	xact_count = cur->xact_count - old->xact_count;
+	server_assignment_count = cur->server_assignment_count - old->server_assignment_count;
 
 	avg->query_count = USEC * query_count / dur;
 	avg->xact_count = USEC * xact_count / dur;
+	avg->server_assignment_count = USEC * server_assignment_count / dur;
 
 	avg->client_bytes = USEC * (cur->client_bytes - old->client_bytes) / dur;
 	avg->server_bytes = USEC * (cur->server_bytes - old->server_bytes) / dur;
@@ -70,18 +86,29 @@ static void calc_average(PgStats *avg, PgStats *cur, PgStats *old)
 	if (xact_count > 0)
 		avg->xact_time = (cur->xact_time - old->xact_time) / xact_count;
 
-	avg->wait_time = USEC * (cur->wait_time - old->wait_time) / dur;
+	if (server_assignment_count > 0)
+		avg->wait_time = (cur->wait_time - old->wait_time) / server_assignment_count;
+
+	ps_client_parse_count = cur->ps_client_parse_count - old->ps_client_parse_count;
+	ps_server_parse_count = cur->ps_server_parse_count - old->ps_server_parse_count;
+	ps_bind_count = cur->ps_bind_count - old->ps_bind_count;
+
+	avg->ps_client_parse_count = USEC * ps_client_parse_count / dur;
+	avg->ps_server_parse_count = USEC * ps_server_parse_count / dur;
+	avg->ps_bind_count = USEC * ps_bind_count / dur;
 }
 
 static void write_stats(PktBuf *buf, PgStats *stat, PgStats *old, char *dbname)
 {
 	PgStats avg;
 	calc_average(&avg, stat, old);
-	pktbuf_write_DataRow(buf, "sNNNNNNNNNNNNNN", dbname,
+	pktbuf_write_DataRow(buf, "sNNNNNNNNNNNNNNNN", dbname,
+			     stat->server_assignment_count,
 			     stat->xact_count, stat->query_count,
 			     stat->client_bytes, stat->server_bytes,
 			     stat->xact_time, stat->query_time,
 			     stat->wait_time,
+			     avg.server_assignment_count,
 			     avg.xact_count, avg.query_count,
 			     avg.client_bytes, avg.server_bytes,
 			     avg.xact_time, avg.query_time,
@@ -93,13 +120,11 @@ bool admin_database_stats(PgSocket *client, struct StatList *pool_list)
 	PgPool *pool;
 	struct List *item;
 	PgDatabase *cur_db = NULL;
-	PgStats st_total, st_db, old_db, old_total;
+	PgStats st_db, old_db;
 	PktBuf *buf;
 
-	reset_stats(&st_total);
 	reset_stats(&st_db);
 	reset_stats(&old_db);
-	reset_stats(&old_total);
 
 	buf = pktbuf_dynamic(512);
 	if (!buf) {
@@ -107,11 +132,13 @@ bool admin_database_stats(PgSocket *client, struct StatList *pool_list)
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "sNNNNNNNNNNNNNN", "database",
+	pktbuf_write_RowDescription(buf, "sNNNNNNNNNNNNNNNN", "database",
+				    "total_server_assignment_count",
 				    "total_xact_count", "total_query_count",
 				    "total_received", "total_sent",
 				    "total_xact_time", "total_query_time",
 				    "total_wait_time",
+				    "avg_server_assignment_count",
 				    "avg_xact_count", "avg_query_count",
 				    "avg_recv", "avg_sent",
 				    "avg_xact_time", "avg_query_time",
@@ -126,8 +153,6 @@ bool admin_database_stats(PgSocket *client, struct StatList *pool_list)
 			write_stats(buf, &st_db, &old_db, cur_db->name);
 
 			cur_db = pool->db;
-			stat_add(&st_total, &st_db);
-			stat_add(&old_total, &old_db);
 			reset_stats(&st_db);
 			reset_stats(&old_db);
 		}
@@ -137,8 +162,6 @@ bool admin_database_stats(PgSocket *client, struct StatList *pool_list)
 	}
 	if (cur_db) {
 		write_stats(buf, &st_db, &old_db, cur_db->name);
-		stat_add(&st_total, &st_db);
-		stat_add(&old_total, &old_db);
 	}
 	admin_flush(client, buf, "SHOW");
 
@@ -149,7 +172,8 @@ static void write_stats_totals(PktBuf *buf, PgStats *stat, PgStats *old, char *d
 {
 	PgStats avg;
 	calc_average(&avg, stat, old);
-	pktbuf_write_DataRow(buf, "sNNNNNNN", dbname,
+	pktbuf_write_DataRow(buf, "sNNNNNNNN", dbname,
+			     stat->server_assignment_count,
 			     stat->xact_count, stat->query_count,
 			     stat->client_bytes, stat->server_bytes,
 			     stat->xact_time, stat->query_time,
@@ -161,13 +185,11 @@ bool admin_database_stats_totals(PgSocket *client, struct StatList *pool_list)
 	PgPool *pool;
 	struct List *item;
 	PgDatabase *cur_db = NULL;
-	PgStats st_total, st_db, old_db, old_total;
+	PgStats st_db, old_db;
 	PktBuf *buf;
 
-	reset_stats(&st_total);
 	reset_stats(&st_db);
 	reset_stats(&old_db);
-	reset_stats(&old_total);
 
 	buf = pktbuf_dynamic(512);
 	if (!buf) {
@@ -175,7 +197,8 @@ bool admin_database_stats_totals(PgSocket *client, struct StatList *pool_list)
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "sNNNNNNN", "database",
+	pktbuf_write_RowDescription(buf, "sNNNNNNNN", "database",
+				    "server_assignment_count",
 				    "xact_count", "query_count",
 				    "bytes_received", "bytes_sent",
 				    "xact_time", "query_time",
@@ -190,8 +213,6 @@ bool admin_database_stats_totals(PgSocket *client, struct StatList *pool_list)
 			write_stats_totals(buf, &st_db, &old_db, cur_db->name);
 
 			cur_db = pool->db;
-			stat_add(&st_total, &st_db);
-			stat_add(&old_total, &old_db);
 			reset_stats(&st_db);
 			reset_stats(&old_db);
 		}
@@ -201,8 +222,6 @@ bool admin_database_stats_totals(PgSocket *client, struct StatList *pool_list)
 	}
 	if (cur_db) {
 		write_stats_totals(buf, &st_db, &old_db, cur_db->name);
-		stat_add(&st_total, &st_db);
-		stat_add(&old_total, &old_db);
 	}
 	admin_flush(client, buf, "SHOW");
 
@@ -213,7 +232,8 @@ static void write_stats_averages(PktBuf *buf, PgStats *stat, PgStats *old, char
 {
 	PgStats avg;
 	calc_average(&avg, stat, old);
-	pktbuf_write_DataRow(buf, "sNNNNNNN", dbname,
+	pktbuf_write_DataRow(buf, "sNNNNNNNN", dbname,
+			     avg.server_assignment_count,
 			     avg.xact_count, avg.query_count,
 			     avg.client_bytes, avg.server_bytes,
 			     avg.xact_time, avg.query_time,
@@ -225,13 +245,11 @@ bool admin_database_stats_averages(PgSocket *client, struct StatList *pool_list)
 	PgPool *pool;
 	struct List *item;
 	PgDatabase *cur_db = NULL;
-	PgStats st_total, st_db, old_db, old_total;
+	PgStats st_db, old_db;
 	PktBuf *buf;
 
-	reset_stats(&st_total);
 	reset_stats(&st_db);
 	reset_stats(&old_db);
-	reset_stats(&old_total);
 
 	buf = pktbuf_dynamic(512);
 	if (!buf) {
@@ -239,7 +257,8 @@ bool admin_database_stats_averages(PgSocket *client, struct StatList *pool_list)
 		return true;
 	}
 
-	pktbuf_write_RowDescription(buf, "sNNNNNNN", "database",
+	pktbuf_write_RowDescription(buf, "sNNNNNNNN", "database",
+				    "server_assignment_count",
 				    "xact_count", "query_count",
 				    "bytes_received", "bytes_sent",
 				    "xact_time", "query_time",
@@ -254,8 +273,6 @@ bool admin_database_stats_averages(PgSocket *client, struct StatList *pool_list)
 			write_stats_averages(buf, &st_db, &old_db, cur_db->name);
 
 			cur_db = pool->db;
-			stat_add(&st_total, &st_db);
-			stat_add(&old_total, &old_db);
 			reset_stats(&st_db);
 			reset_stats(&old_db);
 		}
@@ -265,8 +282,6 @@ bool admin_database_stats_averages(PgSocket *client, struct StatList *pool_list)
 	}
 	if (cur_db) {
 		write_stats_averages(buf, &st_db, &old_db, cur_db->name);
-		stat_add(&st_total, &st_db);
-		stat_add(&old_total, &old_db);
 	}
 	admin_flush(client, buf, "SHOW");
 
@@ -303,6 +318,7 @@ bool show_stat_totals(PgSocket *client, struct StatList *pool_list)
 #define WTOTAL(name) pktbuf_write_DataRow(buf, "sN", "total_" #name, st_total.name)
 #define WAVG(name) pktbuf_write_DataRow(buf, "sN", "avg_" #name, avg.name)
 
+	WTOTAL(server_assignment_count);
 	WTOTAL(xact_count);
 	WTOTAL(query_count);
 	WTOTAL(client_bytes);
@@ -310,6 +326,7 @@ bool show_stat_totals(PgSocket *client, struct StatList *pool_list)
 	WTOTAL(xact_time);
 	WTOTAL(query_time);
 	WTOTAL(wait_time);
+	WAVG(server_assignment_count);
 	WAVG(xact_count);
 	WAVG(query_count);
 	WAVG(client_bytes);
@@ -351,12 +368,19 @@ static void refresh_stats(evutil_socket_t s, short flags, void *arg)
 	if (cf_log_stats) {
 		log_info("stats: %" PRIu64 " xacts/s,"
 			 " %" PRIu64 " queries/s,"
+			 " %" PRIu64 " client parses/s,"
+			 " %" PRIu64 " server parses/s,"
+			 " %" PRIu64 " binds/s,"
 			 " in %" PRIu64 " B/s,"
 			 " out %" PRIu64 " B/s,"
 			 " xact %" PRIu64 " us,"
 			 " query %" PRIu64 " us,"
 			 " wait %" PRIu64 " us",
-			 avg.xact_count, avg.query_count,
+			 avg.xact_count,
+			 avg.query_count,
+			 avg.ps_client_parse_count,
+			 avg.ps_server_parse_count,
+			 avg.ps_bind_count,
 			 avg.client_bytes, avg.server_bytes,
 			 avg.xact_time, avg.query_time,
 			 avg.wait_time);
@@ -365,12 +389,19 @@ static void refresh_stats(evutil_socket_t s, short flags, void *arg)
 	sd_notifyf(0,
 		   "STATUS=stats: %" PRIu64 " xacts/s,"
 		   " %" PRIu64 " queries/s,"
+		   " %" PRIu64 " client parses/s,"
+		   " %" PRIu64 " server parses/s,"
+		   " %" PRIu64 " binds/s,"
 		   " in %" PRIu64 " B/s,"
 		   " out %" PRIu64 " B/s,"
 		   " xact %" PRIu64 " μs,"
 		   " query %" PRIu64 " μs,"
 		   " wait %" PRIu64 " μs",
-		   avg.xact_count, avg.query_count,
+		   avg.xact_count,
+		   avg.query_count,
+		   avg.ps_client_parse_count,
+		   avg.ps_server_parse_count,
+		   avg.ps_bind_count,
 		   avg.client_bytes, avg.server_bytes,
 		   avg.xact_time, avg.query_time,
 		   avg.wait_time);
