@@ -1,10 +1,19 @@
+import getpass
 import re
+import subprocess
 import time
 
 import psycopg
 import pytest
 
-from .utils import LONG_PASSWORD, PG_SUPPORTS_SCRAM, WINDOWS
+from .utils import (
+    FREEBSD,
+    LONG_PASSWORD,
+    MACOS,
+    PG_SUPPORTS_SCRAM,
+    TLS_SUPPORT,
+    WINDOWS,
+)
 
 
 @pytest.mark.md5
@@ -404,6 +413,82 @@ def test_auth_query_database_setting(
                 )
 
 
+@pytest.mark.skipif("WINDOWS", reason="Windows does not have SIGHUP")
+def test_auth_query_works_with_configured_users(bouncer):
+    """
+    Check that when a user is configured with per-user options, but missing from auth_file
+    pgBouncer will still attempt to valididate passwords if auth_query is configured.
+    """
+
+    config = f"""
+        [databases]
+        postgres = host={bouncer.pg.host} port={bouncer.pg.port}
+        [pgbouncer]
+        auth_query = SELECT usename, passwd FROM pg_shadow where usename = $1
+        auth_user = pswcheck
+        stats_users = stats
+        listen_addr = {bouncer.host}
+        admin_users = pgbouncer
+        auth_type = md5
+        auth_file = {bouncer.auth_path}
+        listen_port = {bouncer.port}
+        logfile = {bouncer.log_path}
+        auth_dbname = postgres
+        pool_mode = session
+        [users]
+        puser1 = pool_mode=statement
+    """
+
+    # As a sanity check, make sure that a user with a password in auth_file cannot run transactions
+    # while configured to be in statement pooling mode
+    with bouncer.run_with_config(config):
+        with pytest.raises(psycopg.OperationalError):
+            with bouncer.log_contains(
+                "closing because: transaction blocks not allowed in statement pooling mode"
+            ):
+                bouncer.sql(
+                    query="begin",
+                    user="puser1",
+                    password="foo",
+                    dbname="postgres",
+                )
+
+    config = f"""
+        [databases]
+        postgres = host={bouncer.pg.host} port={bouncer.pg.port}
+        [pgbouncer]
+        auth_query = SELECT usename, passwd FROM pg_shadow where usename = $1
+        auth_user = pswcheck
+        stats_users = stats
+        listen_addr = {bouncer.host}
+        admin_users = pgbouncer
+        auth_type = md5
+        auth_file = {bouncer.auth_path}
+        listen_port = {bouncer.port}
+        logfile = {bouncer.log_path}
+        auth_dbname = postgres
+        pool_mode = session
+        [users]
+        stats = pool_mode=statement
+    """
+
+    # While pgbouncer is set to use session mode by default, the stats user
+    # is set to use statement pooling. pgBouncer should fail to allow a begin
+    # statement while in statement pooling mode, but still be able to authenticate
+    # using auth_query.
+    with bouncer.run_with_config(config):
+        with pytest.raises(psycopg.OperationalError):
+            with bouncer.log_contains(
+                "closing because: transaction blocks not allowed in statement pooling mode"
+            ):
+                bouncer.sql(
+                    query="begin",
+                    user="stats",
+                    password="stats",
+                    dbname="postgres",
+                )
+
+
 @pytest.mark.skipif("WINDOWS", reason="Windows does not have SIGHUP")
 @pytest.mark.md5
 def test_auth_dbname_works_fine(
@@ -576,3 +661,232 @@ async def test_change_server_password_server_lifetime(bouncer, pg):
                 await result3
     finally:
         pg.sql("ALTER USER puser1 PASSWORD 'foo'")
+
+
+@pytest.mark.skipif("MACOS", reason="SSL tests are broken on OSX in CI #1031")
+@pytest.mark.skipif("WINDOWS", reason="Windows does not have SIGHUP")
+@pytest.mark.skipif(not TLS_SUPPORT, reason="pgbouncer is built without TLS support")
+def test_client_hba_cert(bouncer, cert_dir):
+    root = cert_dir / "TestCA1" / "ca.crt"
+    key = cert_dir / "TestCA1" / "sites" / "01-localhost.key"
+    cert = cert_dir / "TestCA1" / "sites" / "01-localhost.crt"
+
+    bouncer.write_ini(f"client_tls_key_file = {key}")
+    bouncer.write_ini(f"client_tls_cert_file = {cert}")
+    bouncer.write_ini(f"client_tls_ca_file = {root}")
+    bouncer.write_ini(f"client_tls_sslmode = require")
+    bouncer.write_ini(f"auth_type = hba")
+    bouncer.write_ini(
+        f"auth_query = SELECT usename, passwd FROM pg_shadow where usename = $1"
+    )
+    bouncer.write_ini(f"auth_user = pswcheck")
+    bouncer.write_ini(f"auth_file = {bouncer.auth_path}")
+    bouncer.write_ini(f"auth_hba_file = pgbouncer_hba.conf")
+    bouncer.write_ini(f"auth_ident_file = pgident.conf")
+
+    bouncer.admin("reload")
+
+    client_key = cert_dir / "TestCA1" / "sites" / "04-pgbouncer.acme.org.key"
+    client_cert = cert_dir / "TestCA1" / "sites" / "04-pgbouncer.acme.org.crt"
+
+    # The client connects to p0x using a client certificate with CN=pgbouncer.acme.org.
+    # hba_eval returns the following line:
+    #    hostssl p0x    all        0.0.0.0/0               cert    map=test
+    # where "test" map is defined in pgident.conf as
+    #    test            pgbouncer.acme.org      someuser
+    #    test            pgbouncer.acme.org      anotheruser
+    # hence the test succeeds.
+    bouncer.psql_test(
+        dbname="p0x",
+        host="localhost",
+        user="someuser",
+        sslmode="verify-full",
+        sslkey=client_key,
+        sslcert=client_cert,
+        sslrootcert=root,
+    )
+
+    bouncer.pg.sql("create user anotheruser with login;")
+
+    # The client connects to p0x using a client certificate with CN=pgbouncer.acme.org.
+    # hba_eval returns the following line:
+    #    hostssl p0x    all        0.0.0.0/0               cert    map=test
+    # where "test" map is defined in pgident.conf as
+    #    test            pgbouncer.acme.org      someuser
+    #    test            pgbouncer.acme.org      anotheruser
+    # hence the test succeeds.
+    bouncer.psql_test(
+        dbname="p0x",
+        host="localhost",
+        user="anotheruser",
+        sslmode="verify-full",
+        sslkey=client_key,
+        sslcert=client_cert,
+        sslrootcert=root,
+    )
+
+    # The client connects to p0x using a client certificate with CN=pgbouncer.acme.org.
+    # hba_eval returns the following line:
+    #    hostssl p0x    all        0.0.0.0/0               cert    map=test
+    # where "test" map is defined in pgident.conf as
+    #    test            pgbouncer.acme.org      someuser
+    #    test            pgbouncer.acme.org      anotheruser
+    # the username 'bouncer' does not match any mapped pg-username.
+    # hence the test fails.
+    with pytest.raises(
+        subprocess.CalledProcessError,
+    ):
+        with bouncer.log_contains(
+            "p0x/bouncer@127.0.0.1:43544 ident map: test does not have a match"
+        ):
+            bouncer.psql_test(
+                dbname="p0x",
+                host="localhost",
+                user="bouncer",
+                sslmode="verify-full",
+                sslkey=client_key,
+                sslcert=client_cert,
+                sslrootcert=root,
+            )
+
+    client_key = cert_dir / "TestCA1" / "sites" / "02-bouncer.key"
+    client_cert = cert_dir / "TestCA1" / "sites" / "02-bouncer.crt"
+
+    # The client connects to p0 using a client certificate with CN=bouncer.
+    # hba_eval returns the following line:
+    #    hostssl p0              bouncer         0.0.0.0/0               cert
+    # CN expected in map is "bouncer" which matches the CN in the client cert
+    # hence the test succeeds.
+    bouncer.psql_test(
+        dbname="p0",
+        host="localhost",
+        user="bouncer",
+        sslmode="verify-full",
+        sslkey=client_key,
+        sslcert=client_cert,
+        sslrootcert=root,
+    )
+
+    # The client connects to p0y using a client certificate with CN=bouncer.
+    # hba_eval returns the following line:
+    #    hostssl p0y             all             0.0.0.0/0               cert    map=test2
+    # where
+    #   test2           bouncer                 all
+    #   test2           pgbouncer.acme.org      "anotheruser"
+    # test2 mapping allows any client with CN="bouncer" to connect using any user name.
+    # Hence the test succeeds.
+    bouncer.psql_test(
+        dbname="p0y",
+        host="localhost",
+        user="someuser",
+        sslmode="verify-full",
+        sslkey=client_key,
+        sslcert=client_cert,
+        sslrootcert=root,
+    )
+
+    client_key = cert_dir / "TestCA1" / "sites" / "04-pgbouncer.acme.org.key"
+    client_cert = cert_dir / "TestCA1" / "sites" / "04-pgbouncer.acme.org.crt"
+
+    # The client connects to p0y using a client certificate with CN=pgbouncer.acme.org.
+    # hba_eval returns the following line:
+    #    hostssl p0y             all             0.0.0.0/0               cert    map=test2
+    # where
+    #   test2           bouncer                 all
+    #   test2           pgbouncer.acme.org      "anotheruser"
+    # for CN=pgbouncer.acme.org, test2 allows to use anotheruser. Hence the test fails.
+
+    with pytest.raises(
+        subprocess.CalledProcessError,
+    ):
+        with bouncer.log_contains(
+            "p0y/someuser@127.0.0.1:39712 ident map: test2 does not have a match"
+        ):
+            bouncer.psql_test(
+                dbname="p0y",
+                host="localhost",
+                user="someuser",
+                sslmode="verify-full",
+                sslkey=client_key,
+                sslcert=client_cert,
+                sslrootcert=root,
+            )
+
+    # The client connects to p0y using a client certificate with CN=pgbouncer.acme.org.
+    # hba_eval returns the following line:
+    #    hostssl p0y             all             0.0.0.0/0               cert    map=test2
+    # where
+    #   test2           bouncer                 all
+    #   test2           pgbouncer.acme.org      "anotheruser"
+    # for CN=pgbouncer.acme.org, test2 allows to use anotheruser. Hence the test succeeds.
+
+    bouncer.psql_test(
+        dbname="p0y",
+        host="localhost",
+        user="anotheruser",
+        sslmode="verify-full",
+        sslkey=client_key,
+        sslcert=client_cert,
+        sslrootcert=root,
+    )
+
+
+@pytest.mark.skipif("WINDOWS", reason="Windows does not have peer authentication")
+def test_peer_auth_ident_map(bouncer):
+    cur_user = getpass.getuser()
+
+    ident_conf_file = bouncer.config_dir / "ident.conf"
+    hba_conf_file = bouncer.config_dir / "hba.conf"
+
+    with open(ident_conf_file, "w") as f:
+        f.write(f"mymap {cur_user} postgres\n")
+        f.write(f"mymap {cur_user} someuser\n")
+
+    with open(hba_conf_file, "w") as f:
+        f.write(f"local   all  all peer map=mymap")
+
+    bouncer.write_ini(f"auth_type = hba")
+    bouncer.write_ini(
+        f"auth_query = SELECT usename, passwd FROM pg_shadow where usename = $1"
+    )
+    bouncer.write_ini(f"auth_user = pswcheck")
+    bouncer.write_ini(f"auth_file = {bouncer.auth_path}")
+    bouncer.write_ini(f"auth_hba_file = {hba_conf_file}")
+    bouncer.write_ini(f"auth_ident_file = {ident_conf_file}")
+
+    bouncer.admin("reload")
+
+    bouncer.psql_test(
+        dbname="p0y",
+        host=f"{bouncer.admin_host}",
+        user="postgres",
+    )
+
+    bouncer.psql_test(
+        dbname="p0y",
+        host=f"{bouncer.admin_host}",
+        user="someuser",
+    )
+
+    with pytest.raises(
+        subprocess.CalledProcessError,
+    ):
+        with bouncer.log_contains(
+            "p0y/bouncer@unix(6202):10202 ident map mymap cannot be matched"
+        ):
+            bouncer.psql_test(
+                dbname="p0y",
+                host=f"{bouncer.admin_host}",
+                user="bouncer",
+            )
+
+    with open(ident_conf_file, "w") as f:
+        f.write(f"mymap {cur_user} all")
+
+    bouncer.admin("reload")
+
+    bouncer.psql_test(
+        dbname="p0",
+        host=f"{bouncer.admin_host}",
+        user="bouncer",
+    )
